/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: eigen-slepc.c,v 1.19 2009/12/01 05:52:17 zlb Exp $ */

#include "phg.h"

#if USE_SLEPC

#include <slepceps.h>
#include <stdlib.h>

static void
setup_petsc_mat(MAT *A, Mat *a)
/* Copy data from MAT A to PETSc Mat a */
{
    MAT_ROW *row;
    PetscInt prealloc = PETSC_DECIDE, *d_nnz, *o_nnz;
    PetscInt I, i, j, d, o, n0, maxcols, *cols;
    PetscScalar *v = NULL;
    INT v_size = 0;

    if (A == NULL) {
	a = PETSC_NULL;
	return;
    }

    assert(sizeof(INT) == sizeof(PetscInt));
    assert(A->type == PHG_UNPACKED);

    d_nnz = phgAlloc(A->rmap->nlocal * sizeof(*d_nnz));
    o_nnz = phgAlloc(A->rmap->nlocal * sizeof(*o_nnz));
    maxcols = 0;
    for (i = 0, row = A->rows; i < A->rmap->nlocal; i++, row++) {
	d = o = 0;
	for (j = 0; j < row->ncols; j++)
	    (row->cols[j] < A->rmap->nlocal) ? d++ : o++;
	d_nnz[i] = d;
	o_nnz[i] = o;
	if (maxcols < row->ncols)
	    maxcols = row->ncols;
    }
    MatCreateMPIAIJ(A->rmap->comm, A->rmap->nlocal, A->cmap->nlocal,
		    A->rmap->nglobal, A->cmap->nglobal,
		    prealloc, d_nnz, prealloc, o_nnz, a);
    MatSetFromOptions(*a);
    phgFree(d_nnz);
    phgFree(o_nnz);

    n0 = I = A->rmap->partition[A->rmap->rank];
    cols = phgAlloc(maxcols * sizeof(*cols));
    for (i = 0, row = A->rows; i < A->rmap->nlocal; i++, I++, row++) {
	if (row->ncols > 0) {
	    for (j = 0; j < row->ncols; j++) {
		d = row->cols[j];
		if (d < A->rmap->nlocal)
		    d += n0;
		else
		    d = A->O2Gmap[d - A->rmap->nlocal];
		cols[j] = d;
	    }
	    if (sizeof(FLOAT) == sizeof(PetscScalar)) {
		MatSetValues(*a, 1, &I, row->ncols, cols,
				(PetscScalar *)row->data, INSERT_VALUES);
	    }
	    else {
		if (v_size < row->ncols) {
		    phgFree(v);
		    v = phgAlloc((v_size = row->ncols) * sizeof(*v));
		}
		for (j = 0; j < row->ncols; j++)
		    v[j] = (PetscScalar)row->data[j];
		MatSetValues(*a, 1, &I, row->ncols, cols, v, INSERT_VALUES);
	    }
	}
    }
    phgFree(cols);
    phgFree(v);

    MatAssemblyBegin(*a, MAT_FINAL_ASSEMBLY);
    MatAssemblyEnd(*a, MAT_FINAL_ASSEMBLY);
}

static PetscScalar	*array;
static EPSWhich		whch;

static int
ecomp(const void *p0, const void *p1)
{
    PetscScalar *v0 = array + *(const int *)p0;
    PetscScalar *v1 = array + *(const int *)p1;
    int ret;

    ret = (*v0 < *v1 ? -1 : (*v0 > *v1 ? 1 : 0));

    return whch == EPS_SMALLEST_MAGNITUDE ? ret : -ret;
}

EigenSolverPrototype(phgEigenSolverSLEPc)
/* (A, B, n, which, tau, evals, evecs, eps, itmax, nit) */
{
    static BOOLEAN	initialized = FALSE;
    VEC			*vec;
    FLOAT		*data;
    int			i, j, nev, ncv, its, *ind;
    Mat			mA, mB;		/* matrices */
    Vec			evec;
    EPS			solver;		/* eigenproblem solver context */
    PetscScalar		eval;
    PetscErrorCode	ierr;

    if (!initialized) {
	/* register options */
	initialized = TRUE;
	/*phgOptionsRegisterTitle("\nSLEPc options:", "\n", "slepc");*/
	return 0;
    }

    if (n <= 0 || A->cmap->nglobal <= 0)
	return 0;

    vec = *evecs;
    MagicCheck(VEC, vec);

    if (n > A->cmap->nglobal)
	n = A->cmap->nglobal;

    if (vec == NULL) {
	*evecs = vec = phgMapCreateVec(A->cmap, n);
    }
    else {
	if (vec->nvec < n) {
	    phgFree(vec->data);
	    vec->data = phgAlloc(n * A->cmap->nlocal * sizeof(*vec->data));
	}
	vec->nvec = n;
	/* TODO: set initial vector */
    }

    assert(A->rmap->nglobal > 1);

    setup_petsc_mat(A, &mA);
    setup_petsc_mat(B, &mB);

    ierr = EPSCreate(A->cmap->comm, &solver);
    ierr = EPSSetOperators(solver, mA, mB);
    /* FIXME: why EPS_GHEP does not work! */
    /*ierr = EPSSetProblemType(solver, EPS_GHEP);*/
    ierr = EPSSetFromOptions(solver);
    ierr = EPSSetTolerances(solver, eps, itmax * n);
    ncv = 2 * n;
#if SLEPC_VERSION_MAJOR < 3
    ierr = EPSSetDimensions(solver, n, ncv);
#else	/* SLEPC_VERSION_MAJOR < 3 */
    ierr = EPSSetDimensions(solver, n, ncv, 0);
#endif	/* SLEPC_VERSION_MAJOR < 3 */
    whch = (which <= 0 ? EPS_SMALLEST_MAGNITUDE : EPS_LARGEST_MAGNITUDE);
    ierr = EPSSetWhichEigenpairs(solver, whch);
    ierr = EPSSolve(solver);
    ierr = EPSGetConverged(solver, &nev);
    ind = phgAlloc(nev * sizeof(*ind));
    if (nev > 0) {
	array = phgAlloc(nev * sizeof(*array));
	for (i = 0; i < nev; i++) {
	    ind[i] = i;
	    EPSGetEigenpair(solver, i, array + i, PETSC_NULL,
					PETSC_NULL, PETSC_NULL);
	}
#if 0
	EPSSortEigenvalues(nev, array, array, whch, n, ind);
#else
	qsort(ind, nev, sizeof(*ind), ecomp);
#endif
	phgFree(array);
    }
    if (nev > n)
	nev = n;
    vec->nvec = nev;
    if (nev == 0) {
	phgFree(vec->data);
	vec->data = NULL;
    }
    ierr = EPSGetIterationNumber(solver, &its);
    *nit = its;

    ierr = MatGetVecs(mA, PETSC_NULL, &evec);
    data = vec->data;
#if DEBUG
    VecGetLocalSize(evec, &j);
    assert(j == A->cmap->nlocal);
#endif	/* DEBUG */
    for (i = 0; i < nev; i++) {
	/*ierr = EPSComputeRelativeError(solver, i, &error);*/
	EPSGetEigenpair(solver, ind[i], &eval, PETSC_NULL, evec, PETSC_NULL);
	evals[i] = eval;
	VecGetArray(evec, &array);
	for (j = 0; j < A->cmap->nlocal; j++)
	    *(data++) = array[j];
	VecRestoreArray(evec, &array);
    }
    phgFree(ind);

    ierr = EPSDestroy(solver);
    ierr = VecDestroy(evec);
    ierr = MatDestroy(mA);
    if (mB != NULL)
	ierr = MatDestroy(mB);

    return nev;
}

#else	/* USE_SLEPC */
EigenSolverVoid(phgEigenSolverSLEPc)
#endif	/* USE_SLEPC */
